# Copyright 2020-present, Mayo Clinic Department of Neurology - Laboratory of Bioelectronics Neurophysiology and Engineering
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


"""
Python Invasive Electrophysiology Signal Processing Toolbox - sleep_classification (c) is a Python package developed by the `Bioelectronics Neurophysiology and Engineering <https://www.mayo.edu/research/labs/bioelectronics-neurophysiology-engineering/overview>`_ laboratory of Department of Neurology Mayo Clinic.
The main scope of this package is to provide tools for an automated sleep classification using intracranial EEG signals.
"""

__version__ = '0.1.5'



import os
# Check windows or linux and sets separator
if os.name == 'nt': DELIMITER = '\\'
else: DELIMITER = '/'









