# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.




"""
A package for a standardized manipulation with hypnogram annotations, saving, loading across different file formats (CyberP{SG, NSRR, ...) but also
basic analysis of hypnograms. The package is based on pandas.DataFrame.
"""














